--
library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.std_logic_arith.all;
use IEEE.std_logic_unsigned.all;

entity VGAOUT is
	port (
		DISPD		: in  std_logic_vector(3 downto 0);
		DISPTMG_HS	: in  std_logic;
		DISPTMG_VS	: in  std_logic;
		CLKENB		: in  std_logic;
		SRMODE		: in  std_logic;
		SC4COLORON	: in  std_logic;
		SC4COLORMD	: in  std_logic_vector(1 downto 0);
		RAMDTI		: in  std_logic_vector(15 downto 0);
		DT_LOCK3	: in  std_logic;
		DT_LOCK4	: in  std_logic;
		LOCK_PLL	: in  std_logic;
		CLK108M		: in  std_logic;
		RSTN		: in  std_logic;
		RAMADD		: out std_logic_vector(14 downto 0);
		RAMDTO		: out std_logic_vector(15 downto 0);
		RAMWRN		: out std_logic;
		RAMOEN		: out std_logic;
		RAMCTRLN	: out std_logic;
		VGA_R		: out std_logic_vector(3 downto 0);
		VGA_G		: out std_logic_vector(3 downto 0);
		VGA_B		: out std_logic_vector(3 downto 0);
		VGA_HS		: out std_logic;
		VGA_VS		: out std_logic
	);
end VGAOUT;

architecture RTL of VGAOUT is

	signal init_f1		: std_logic;
	signal init_f2		: std_logic;
	signal init_f3		: std_logic;

	signal hcntwrenb	: std_logic;
	signal hsync_f1		: std_logic;
	signal hcntwr		: std_logic_vector(9 downto 0);
	signal vcntwr		: std_logic_vector(8 downto 0);
	signal w_hvld_i		: std_logic;

	signal vsyncwr_f1	: std_logic;
	signal vsyncwr_f2	: std_logic;

	signal ramwad		: std_logic_vector(16 downto 0);
	signal ramwdt		: std_logic_vector(3 downto 0);
	signal ramwlat		: std_logic_vector(15 downto 0);
	signal ramwlat_f0	: std_logic_vector(15 downto 0);
	signal ramwlat_f1	: std_logic_vector(15 downto 0);
	signal ramwad_f1	: std_logic_vector(16 downto 0);
	signal ramwp		: std_logic;
	signal ramwflag		: std_logic;
	signal ramwflag_f1	: std_logic;
	signal ramwrn_f0	: std_logic;
	signal ramwrn_f1	: std_logic;
	signal ramctrln_f0	: std_logic;
	signal ramctrln_f1	: std_logic;
	signal ramwadout_f0	: std_logic_vector(16 downto 0);
	signal ramwadout_f1	: std_logic_vector(16 downto 0);
	signal ramrenb		: std_logic;
	signal ramrvenb		: std_logic;
	signal ramrad		: std_logic_vector(16 downto 0);
	signal ramrlat_f0	: std_logic_vector(15 downto 0);
	signal ramrlat_f1	: std_logic_vector(15 downto 0);
	signal ramrlat_f2	: std_logic_vector(15 downto 0);
	signal ramrdt		: std_logic_vector(3 downto 0);
	signal ramvcnt		: std_logic_vector(2 downto 0);
	signal ramadd_f0	: std_logic_vector(14 downto 0);
	signal ramadd_f1	: std_logic_vector(14 downto 0);
	signal ramoen_f0	: std_logic;
	signal ramoen_f1	: std_logic;

	signal hcntrdenb	: std_logic;
	signal hcntrd		: std_logic_vector(10 downto 0);
	signal hcntrdtmp	: std_logic_vector(10 downto 0);
	signal vcntrd		: std_logic_vector(10 downto 0);

	signal vga_vs_i		: std_logic;

	signal vga_hs_i		: std_logic;
	signal r_hvld_i		: std_logic;
	signal r_hvld_f1	: std_logic;
	signal r_hvld_f2	: std_logic;
	signal r_hvld2_f1	: std_logic;
	signal r_hvld2_f2	: std_logic;
	signal r_hvld2_f3	: std_logic;
	signal r_hvld2_f4	: std_logic;
	signal r_hvld2_f5	: std_logic;
	signal r_hvld2_f6	: std_logic;
	signal r_hvld2_f7	: std_logic;
	signal r_hvld2_f8	: std_logic;
	signal r_hvld3_f1	: std_logic;
	signal r_hvld3_f2	: std_logic;
	signal r_hvld2_sel	: std_logic;

	signal ramrdt_mux			: std_logic_vector(3 downto 0);
	signal ramrdt_mux_tmp		: std_logic_vector(3 downto 0);

	signal ramrdt_mux_f1	: std_logic_vector(3 downto 0);
	signal ramrdt_mux_f2	: std_logic_vector(3 downto 0);
	signal ramrdt_mux_f3	: std_logic_vector(3 downto 0);
	signal ramrdt_mux_f4	: std_logic_vector(3 downto 0);
	signal ramrdt_mux_f5	: std_logic_vector(3 downto 0);
	signal ramrdt_mux_f6	: std_logic_vector(3 downto 0);
	signal ramrdt_mux_f7	: std_logic_vector(3 downto 0);
	signal ramrdt_mux_f8	: std_logic_vector(3 downto 0);
	signal ramrdt_mux_sel	: std_logic_vector(3 downto 0);

	signal col_b_f2		: std_logic;
	signal col_b_f3		: std_logic;
	signal col_b_f4		: std_logic;
	signal col_b_f5		: std_logic;
	signal col_b_f6		: std_logic;
	signal col_b_f7		: std_logic;
	signal col_b_f8		: std_logic;
	signal col_b_f9		: std_logic;
	signal col_b_f10	: std_logic;
	signal col_w_f2		: std_logic;
	signal col_w_f3		: std_logic;
	signal col_w_f4		: std_logic;
	signal col_w_f5		: std_logic;
	signal col_w_f6		: std_logic;
	signal col_w_f7		: std_logic;
	signal col_w_f8		: std_logic;
	signal col_w_f9		: std_logic;
	signal col_w_f10	: std_logic;

	signal bitpat		: std_logic_vector(4 downto 0);
	signal bitpat_f1	: std_logic_vector(4 downto 0);
	signal bitpatsel	: std_logic_vector(4 downto 0);
	signal allsc4		: std_logic;
	signal allsc4_f1	: std_logic;
	signal allsc4sel	: std_logic;
	signal set2p		: std_logic_vector(23 downto 0);
	signal set2p_f1		: std_logic_vector(23 downto 0);
	signal set2p_f2		: std_logic_vector(23 downto 0);
	signal set2r		: std_logic_vector(23 downto 0);
	signal set2r_f1		: std_logic_vector(23 downto 0);
	signal set2r_f2		: std_logic_vector(23 downto 0);
	signal set2g		: std_logic_vector(23 downto 0);
	signal set2g_f1		: std_logic_vector(23 downto 0);
	signal set2g_f2		: std_logic_vector(23 downto 0);
	signal set2b		: std_logic_vector(23 downto 0);
	signal set2b_f1		: std_logic_vector(23 downto 0);
	signal set2b_f2		: std_logic_vector(23 downto 0);

	signal c_dg			: std_logic_vector(23 downto 0);
	signal c_g			: std_logic_vector(23 downto 0);
	signal c_dp			: std_logic_vector(23 downto 0);
	signal c_p			: std_logic_vector(23 downto 0);
	signal c_db			: std_logic_vector(23 downto 0);
	signal c_b			: std_logic_vector(23 downto 0);
	signal c_dr			: std_logic_vector(23 downto 0);
	signal c_r			: std_logic_vector(23 downto 0);
	signal c_w			: std_logic_vector(23 downto 0);
	signal c_00			: std_logic_vector(23 downto 0);


	signal vgadat		: std_logic_vector(23 downto 0);

	signal vga_hs_f1	: std_logic;
	signal vga_hs_f2	: std_logic;
	signal vga_hs_f3	: std_logic;
	signal vga_hs_f4	: std_logic;
	signal vga_hs_f5	: std_logic;
	signal vga_hs_f6	: std_logic;
	signal vga_hs_f7	: std_logic;
	signal vga_hs_f8	: std_logic;
	signal vga_hs_f9	: std_logic;
	signal vga_hs_f10	: std_logic;
	signal vga_hs_sel	: std_logic;
	signal vga_vs_f1	: std_logic;
	signal vga_vs_f2	: std_logic;
	signal vga_vs_f3	: std_logic;
	signal vga_vs_f4	: std_logic;
	signal vga_vs_f5	: std_logic;
	signal vga_vs_f6	: std_logic;
	signal vga_vs_f7	: std_logic;
	signal vga_vs_f8	: std_logic;
	signal vga_vs_f9	: std_logic;
	signal vga_vs_f10	: std_logic;
	signal vga_vs_sel	: std_logic;

	signal vga_r_o		: std_logic_vector(7 downto 0);
	signal vga_g_o		: std_logic_vector(7 downto 0);
	signal vga_b_o		: std_logic_vector(7 downto 0);
	signal vga_hs_o		: std_logic;
	signal vga_vs_o		: std_logic;


begin

-- ݑi14.31818MHzjPC912NbNi63.6952nsj
-- ǂݏoi108MHzjSC6752NbNi62.3262nsj
-- ǂݏo̕

-- mk2/66̓
-- horizontal
-- iHSYNC H->L j
-- SYNCpXF64clk
-- f[^擪F192clk
-- f[^IF832clk
--
-- vartical
-- iVSYNC H->L j
-- ŚF262rowi#0`#261j
-- VSYNCpXF3rowi#0`#2j
-- f[^F200rowi#33`#232j


-- mk2SR/66SR̓
-- horizontal
-- iHSYNC H->L j
-- SYNCpXF68clk
-- f[^擪F200clk
-- f[^IF840clk
--
-- vartical
-- iVSYNC H->L j
-- ŚF262rowi#0`#261j
-- VSYNCpXF4rowi#0`#3j
-- f[^F200rowi#39`#238j
-- f[^F204rowi#37`#240j
--


---- MAIN reset
	process (CLK108M,RSTN)
	begin
		if (RSTN = '0') then
			init_f1 <= '1';
			init_f2 <= '1';
			init_f3 <= '1';
		elsif (CLK108M'event and CLK108M = '1') then
			if (LOCK_PLL = '1') then
				init_f1 <= '0';
				init_f2 <= init_f1;
				init_f3 <= init_f2;
			end if;
		end if;
	end process;


---- CLK108M write
	process (CLK108M,RSTN)
	begin
		if (RSTN = '0') then
			hcntwrenb <= '0';
		elsif (CLK108M'event and CLK108M = '1') then
			hcntwrenb <= not init_f3;
		end if;
	end process;


---- hsync latch
	process (CLK108M,RSTN)
	begin
		if (RSTN = '0') then
			hsync_f1 <= '1';
		elsif (CLK108M'event and CLK108M = '1') then
			if (CLKENB = '1') then
				hsync_f1 <= DISPTMG_HS;
			end if;
		end if;
	end process;

	process (CLK108M,RSTN)
	begin
		if (RSTN = '0') then
			hcntwr <= (others => '1');
		elsif (CLK108M'event and CLK108M = '1') then
			if (hcntwrenb = '0') then
				hcntwr <= (others => '1');
			elsif (CLKENB = '1') then
				if (DISPTMG_HS = '0' and hsync_f1 = '1') then
					hcntwr <= (others => '0');
				elsif (hcntwr /= "1111111111") then
					hcntwr <= hcntwr + 1;
				end if;
			end if;
		end if;
	end process;


	process (CLK108M,RSTN)
	begin
		if (RSTN = '0') then
			vsyncwr_f1 <= '1';
			vsyncwr_f2 <= '1';
		elsif (CLK108M'event and CLK108M = '1') then
			if (hcntwrenb = '0') then
				vsyncwr_f1 <= '1';
				vsyncwr_f2 <= '1';
			elsif (CLKENB = '1') then
				if (hcntwr = 20) then
					vsyncwr_f1 <= DISPTMG_VS;
					vsyncwr_f2 <= vsyncwr_f1;
				end if;
			end if;
		end if;
	end process;

	process (CLK108M,RSTN)
	begin
		if (RSTN = '0') then
			vcntwr <= (others => '0');
		elsif (CLK108M'event and CLK108M = '1') then
			if (hcntwrenb = '0') then
				vcntwr <= (others => '0');
			elsif (CLKENB = '1') then
				if (hcntwr = 24) then
					if (vsyncwr_f1 = '0' and vsyncwr_f2 = '1') then
						vcntwr <= (others => '0');
					else
						vcntwr <= vcntwr + 1;
					end if;
				end if;
			end if;
		end if;
	end process;

	w_hvld_i <=
		'1' when (SRMODE = '0' and 192 <= hcntwr and hcntwr < 832 and 33 <= vcntwr and vcntwr < 233) else
		'1' when (SRMODE = '1' and 200 <= hcntwr and hcntwr < 840 and 37 <= vcntwr and vcntwr < 241) else
		'0';


	process (CLK108M,RSTN)
	begin
		if (RSTN = '0') then
			ramwad <= (others => '0');
		elsif (CLK108M'event and CLK108M = '1') then
			if (hcntwrenb = '0') then
				ramwad <= (others => '0');
			elsif (CLKENB = '1' and hcntwr = 24 and vsyncwr_f1 = '0' and vsyncwr_f2 = '1') then
				ramwad <= (others => '0');
			elsif (w_hvld_i = '1' and CLKENB = '1') then
				ramwad <= ramwad + 1;
			end if;
		end if;
	end process;

	process (CLK108M,RSTN)
	begin
		if (RSTN = '0') then
			ramwdt <= (others => '0');
		elsif (CLK108M'event and CLK108M = '1') then
			if (CLKENB = '1') then
				ramwdt <= DISPD;
			end if;
		end if;
	end process;

	process (CLK108M,RSTN)
	begin
		if (RSTN = '0') then
			ramwlat <= (others => '0');
		elsif (CLK108M'event and CLK108M = '1') then
			if (w_hvld_i = '1' and CLKENB = '1') then
				if    (ramwad(1 downto 0) = "00") then
					ramwlat( 3 downto  0) <= ramwdt;
				elsif (ramwad(1 downto 0) = "01") then
					ramwlat( 7 downto  4) <= ramwdt;
				elsif (ramwad(1 downto 0) = "10") then
					ramwlat(11 downto  8) <= ramwdt;
				else
					ramwlat(15 downto 12) <= ramwdt;
				end if;
			end if;
		end if;
	end process;

	process (CLK108M,RSTN)
	begin
		if (RSTN = '0') then
			ramwp <= '0';
		elsif (CLK108M'event and CLK108M = '1') then
			if (w_hvld_i = '1' and CLKENB = '1' and ramwad(1 downto 0) = "11") then
				ramwp <= '1';
			else
				ramwp <= '0';
			end if;
		end if;
	end process;

	process (CLK108M,RSTN)
	begin
		if (RSTN = '0') then
			ramwlat_f0 <= (others => '0');
			ramwad_f1  <= (others => '0');
		elsif (CLK108M'event and CLK108M = '1') then
			if (CLKENB = '1' and hcntwr = 24 and vsyncwr_f1 = '0' and vsyncwr_f2 = '1') then
				ramwlat_f0 <= (others => '0');
				ramwad_f1  <= (others => '0');
			elsif (ramwp = '1') then
				ramwlat_f0 <= ramwlat;
				ramwad_f1  <= ramwad;
			end if;
		end if;
	end process;

	process (CLK108M,RSTN)
	begin
		if (RSTN = '0') then
			ramwflag <= '0';
		elsif (CLK108M'event and CLK108M = '1') then
			if (ramwp = '1') then
				ramwflag <= '1';
			elsif (hcntrd(2 downto 0) = "101") then
				ramwflag <= '0';
			end if;
		end if;
	end process;


-- RAM access
-- write(4clk)/read(4clk)
--  _   _   _   _   _   _   _   _
-- | |_| |_| |_| |_| |_| |_| |_| |_|
--
-- |     WAD       |     RAD       |
-- 
-- |     WDT       |    latch  v   |
-- ________________             ___
--                 |___________|     OEN
-- ____     _______________________
--     |___|                         WEN
--              ___________________
-- |___________|                     CTRLN


	process (CLK108M,RSTN)
	begin
		if (RSTN = '0') then
			ramwflag_f1 <= '0';
		elsif (CLK108M'event and CLK108M = '1') then
			if (hcntrd(2 downto 0) = "101") then
				ramwflag_f1 <= ramwflag;
			end if;
		end if;
	end process;

	process (CLK108M,RSTN)
	begin
		if (RSTN = '0') then
			ramwrn_f0 <= '1';
		elsif (CLK108M'event and CLK108M = '1') then
			if (hcntrd(2 downto 0) = "111") then
				ramwrn_f0 <= not ramwflag_f1;
			else
				ramwrn_f0 <= '1';
			end if;
		end if;
	end process;

	process (CLK108M,RSTN)
	begin
		if (RSTN = '0') then
			ramctrln_f0 <= '1';
		elsif (CLK108M'event and CLK108M = '1') then
			if (hcntrd(2 downto 0) = "110") then
				ramctrln_f0 <= not ramwflag_f1;
			elsif (hcntrd(2 downto 0) = "001") then
				ramctrln_f0 <= '1';
			end if;
		end if;
	end process;

	process (CLK108M,RSTN)
	begin
		if (RSTN = '0') then
			ramwrn_f1   <= '1';
			ramctrln_f1 <= '1';
		elsif (CLK108M'event and CLK108M = '1') then
			ramwrn_f1   <= ramwrn_f0;
			ramctrln_f1 <= ramctrln_f0;
		end if;
	end process;

	RAMWRN   <= ramwrn_f1;
	RAMCTRLN <= ramctrln_f1;


	process (CLK108M,RSTN)
	begin
		if (RSTN = '0') then
			ramwlat_f1 <= (others => '0');
		elsif (CLK108M'event and CLK108M = '1') then
			ramwlat_f1 <= ramwlat_f0;
		end if;
	end process;

	RAMDTO <= ramwlat_f1;


	process (CLK108M,RSTN)
	begin
		if (RSTN = '0') then
			ramwadout_f0 <= (others => '0');
			ramwadout_f1 <= (others => '0');
		elsif (CLK108M'event and CLK108M = '1') then
			if (hcntrd(2 downto 0) = "101") then
				ramwadout_f0 <= ramwad_f1;
			end if;
			ramwadout_f1 <= ramwadout_f0;
		end if;
	end process;

	process (CLK108M,RSTN)
	begin
		if (RSTN = '0') then
			ramadd_f0 <= (others => '0');
		elsif (CLK108M'event and CLK108M = '1') then
			if (hcntrd(2 downto 0) = "110") then
				ramadd_f0 <= ramwadout_f1(16 downto 2);
			elsif (hcntrd(2 downto 0) = "010") then
				ramadd_f0 <= ramrad(16 downto 2);
			end if;
		end if;
	end process;

	process (CLK108M,RSTN)
	begin
		if (RSTN = '0') then
			ramadd_f1 <= (others => '0');
		elsif (CLK108M'event and CLK108M = '1') then
			ramadd_f1 <= ramadd_f0;
		end if;
	end process;

	RAMADD <= ramadd_f1;


	process (CLK108M,RSTN)
	begin
		if (RSTN = '0') then
			ramrlat_f0 <= (others => '0');
		elsif (CLK108M'event and CLK108M = '1') then
			ramrlat_f0 <= RAMDTI;
		end if;
	end process;

	process (CLK108M,RSTN)
	begin
		if (RSTN = '0') then
			ramrlat_f1 <= (others => '0');
		elsif (CLK108M'event and CLK108M = '1') then
			if (hcntrd(2 downto 0) = "111") then
				ramrlat_f1 <= ramrlat_f0;
			end if;
		end if;
	end process;

	process (CLK108M,RSTN)
	begin
		if (RSTN = '0') then
			ramrlat_f2 <= (others => '0');
		elsif (CLK108M'event and CLK108M = '1') then
			if (hcntrd(2 downto 0) = "000") then
				ramrlat_f2 <= ramrlat_f1;
			end if;
		end if;
	end process;

	process (CLK108M,RSTN)
	begin
		if (RSTN = '0') then
			ramrdt <= (others => '0');
		elsif (CLK108M'event and CLK108M = '1') then
			if    (hcntrd(2 downto 0) = "001") then
				ramrdt <= ramrlat_f2( 3 downto  0);
			elsif (hcntrd(2 downto 0) = "011") then
				ramrdt <= ramrlat_f2( 7 downto  4);
			elsif (hcntrd(2 downto 0) = "101") then
				ramrdt <= ramrlat_f2(11 downto  8);
			elsif (hcntrd(2 downto 0) = "111") then
				ramrdt <= ramrlat_f2(15 downto 12);
			end if;
		end if;
	end process;


	process (CLK108M,RSTN)
	begin
		if (RSTN = '0') then
			ramoen_f0 <= '1';
		elsif (CLK108M'event and CLK108M = '1') then
			if (ramrenb = '1') then
				if (hcntrd(2 downto 0) = "010") then
					ramoen_f0 <= '0';
				elsif (hcntrd(2 downto 0) = "101") then
					ramoen_f0 <= '1';
				end if;
			else
				ramoen_f0 <= '1';
			end if;
		end if;
	end process;

	process (CLK108M,RSTN)
	begin
		if (RSTN = '0') then
			ramoen_f1 <= '1';
		elsif (CLK108M'event and CLK108M = '1') then
			ramoen_f1 <= ramoen_f0;
		end if;
	end process;

	RAMOEN <= ramoen_f1;


---- CLK108M read
	process (CLK108M,RSTN)
	begin
		if (RSTN = '0') then
			hcntrdenb <= '0';
		elsif (CLK108M'event and CLK108M = '1') then
			hcntrdenb <= not init_f3;
		end if;
	end process;

	process (CLK108M,RSTN)
	begin
		if (RSTN = '0') then
			hcntrd <= (others => '1');
		elsif (CLK108M'event and CLK108M = '1') then
			if (hcntrdenb = '0') then
				hcntrd <= (others => '1');
			elsif (hcntrd = 1688-1) then
				hcntrd <= (others => '0');
			else
				hcntrd <= hcntrd + 1;
			end if;
		end if;
	end process;

	hcntrdtmp <= hcntrd - (112+248);


	process (CLK108M,RSTN)
	begin
		if (RSTN = '0') then
			vcntrd  <= (others => '1');
		elsif (CLK108M'event and CLK108M = '1') then
			if (hcntrdenb = '0') then
				vcntrd  <= (others => '1');
			elsif (hcntrd = 1688-1) then
				if (vcntrd = 1066-1) then
					vcntrd  <= (others => '0');
				else
					vcntrd <= vcntrd + 1;
				end if;
			end if;
		end if;
	end process;


-- horizontal
-- sync=112 / back porch=248 / pixel=1280 / frontporch=48
-- all=1688
--
-- vartical
-- sync=3 / back porch=38 / pixel=1024 (4+204*5+4) / front porch=1
-- all=1066
--

	vga_hs_i <=	'0' when (0 <= hcntrd and hcntrd <   112) else '1';
	vga_vs_i <= '0' when (0 <= vcntrd and vcntrd <     3) else '1';

	r_hvld_i <=
		'1' when (SRMODE = '0' and
			(112+248) <= hcntrd and hcntrd < (112+248+1280) and
			(3+38+12) <= vcntrd and vcntrd < (3+38+12+1000) ) else

		'1' when (SRMODE = '1' and
			(112+248) <= hcntrd and hcntrd < (112+248+1280) and
			(3+38+4)  <= vcntrd and vcntrd < (3+38+4+1020) ) else

		'0';

	ramrenb <=
		'1' when (SRMODE = '0' and
			(112+248-8) <= hcntrd and hcntrd < (112+248+1280-8) and
			(3+38+12)   <= vcntrd and vcntrd < (3+38+12+1000) ) else

		'1' when (SRMODE = '1' and
			(112+248-8) <= hcntrd and hcntrd < (112+248+1280-8) and
			(3+38+4)    <= vcntrd and vcntrd < (3+38+4+1020) ) else

		'0';

	ramrvenb <=
		'1' when (SRMODE = '0' and
			(3+38+12)    <= vcntrd and vcntrd < (3+38+12+1000) ) else

		'1' when (SRMODE = '1' and
			(3+38+4)     <= vcntrd and vcntrd < (3+38+4+1020) ) else

		'0';


	process (CLK108M,RSTN)
	begin
		if (RSTN = '0') then
			ramrad  <= (others => '0');
			ramvcnt <= (others => '0');
		elsif (CLK108M'event and CLK108M = '1') then
			if (hcntrdenb = '0') then
				ramrad  <= (others => '0');
				ramvcnt <= (others => '0');
			elsif (hcntrd = 1688-1) then
				if (vcntrd = 1066-1) then
					ramrad  <= (others => '0');
					ramvcnt <= (others => '0');
				elsif (ramrvenb = '1') then
					if (ramvcnt = "100") then
						ramvcnt <= (others => '0');
					else
						ramvcnt <= ramvcnt + 1;
						ramrad  <= ramrad - 640;
					end if;
				end if;
			elsif (ramrenb = '1' and hcntrd(0) = '1') then
				ramrad <= ramrad + 1;
			end if;
		end if;
	end process;




	process (CLK108M,RSTN)
	begin
		if (RSTN = '0') then
			vga_hs_f1  <= '1';
			vga_hs_f2  <= '1';
			vga_hs_f3  <= '1';
			vga_hs_f4  <= '1';
			vga_hs_f5  <= '1';
			vga_hs_f6  <= '1';
			vga_hs_f7  <= '1';
			vga_hs_f8  <= '1';
			vga_hs_f9  <= '1';
			vga_hs_f10 <= '1';
			vga_vs_f1  <= '1';
			vga_vs_f2  <= '1';
			vga_vs_f3  <= '1';
			vga_vs_f4  <= '1';
			vga_vs_f5  <= '1';
			vga_vs_f6  <= '1';
			vga_vs_f7  <= '1';
			vga_vs_f8  <= '1';
			vga_vs_f9  <= '1';
			vga_vs_f10 <= '1';
		elsif (CLK108M'event and CLK108M = '1') then
			if (hcntrdtmp(0 downto 0) = "1") then
				vga_hs_f1  <= vga_hs_i;
				vga_hs_f2  <= vga_hs_f1;
				vga_hs_f3  <= vga_hs_f2;
				vga_hs_f4  <= vga_hs_f3;
				vga_hs_f5  <= vga_hs_f4;
				vga_hs_f6  <= vga_hs_f5;
				vga_hs_f7  <= vga_hs_f6;
				vga_hs_f8  <= vga_hs_f7;
				vga_hs_f9  <= vga_hs_f8;
				vga_hs_f10 <= vga_hs_f9;
				vga_vs_f1  <= vga_vs_i;
				vga_vs_f2  <= vga_vs_f1;
				vga_vs_f3  <= vga_vs_f2;
				vga_vs_f4  <= vga_vs_f3;
				vga_vs_f5  <= vga_vs_f4;
				vga_vs_f6  <= vga_vs_f5;
				vga_vs_f7  <= vga_vs_f6;
				vga_vs_f8  <= vga_vs_f7;
				vga_vs_f9  <= vga_vs_f8;
				vga_vs_f10 <= vga_vs_f9;
			end if;
		end if;
	end process;


	process (CLK108M,RSTN)
	begin
		if (RSTN = '0') then
			r_hvld_f1  <= '0';
			r_hvld_f2  <= '0';
		elsif (CLK108M'event and CLK108M = '1') then
			r_hvld_f1  <= r_hvld_i;
			r_hvld_f2  <= r_hvld_f1;
		end if;
	end process;

	ramrdt_mux_tmp <= ramrdt when (r_hvld_f2 = '1') else (others => '0');

	process (CLK108M,RSTN)
	begin
		if (RSTN = '0') then
			ramrdt_mux <= (others => '0');
		elsif (CLK108M'event and CLK108M = '1') then
			if (hcntrdtmp(0 downto 0) = "1") then
				ramrdt_mux <= ramrdt_mux_tmp;
			end if;
		end if;
	end process;

	process (CLK108M,RSTN)
	begin
		if (RSTN = '0') then
			ramrdt_mux_f1  <= (others => '0');
			ramrdt_mux_f2  <= (others => '0');
			ramrdt_mux_f3  <= (others => '0');
			ramrdt_mux_f4  <= (others => '0');
			ramrdt_mux_f5  <= (others => '0');
			ramrdt_mux_f6  <= (others => '0');
			ramrdt_mux_f7  <= (others => '0');
			ramrdt_mux_f8  <= (others => '0');
			col_b_f2       <= '0';
			col_b_f3       <= '0';
			col_b_f4       <= '0';
			col_b_f5       <= '0';
			col_b_f6       <= '0';
			col_b_f7       <= '0';
			col_b_f8       <= '0';
			col_b_f9       <= '0';
			col_b_f10      <= '0';
			col_w_f2       <= '0';
			col_w_f3       <= '0';
			col_w_f4       <= '0';
			col_w_f5       <= '0';
			col_w_f6       <= '0';
			col_w_f7       <= '0';
			col_w_f8       <= '0';
			col_w_f9       <= '0';
			col_w_f10      <= '0';
			r_hvld2_f1     <= '0';
			r_hvld2_f2     <= '0';
			r_hvld2_f3     <= '0';
			r_hvld2_f4     <= '0';
			r_hvld2_f5     <= '0';
			r_hvld2_f6     <= '0';
			r_hvld2_f7     <= '0';
			r_hvld2_f8     <= '0';
		elsif (CLK108M'event and CLK108M = '1') then
			if (hcntrdtmp(0 downto 0) = "1") then

				ramrdt_mux_f1  <= ramrdt_mux;
				ramrdt_mux_f2  <= ramrdt_mux_f1;
				ramrdt_mux_f3  <= ramrdt_mux_f2;
				ramrdt_mux_f4  <= ramrdt_mux_f3;
				ramrdt_mux_f5  <= ramrdt_mux_f4;
				ramrdt_mux_f6  <= ramrdt_mux_f5;
				ramrdt_mux_f7  <= ramrdt_mux_f6;
				ramrdt_mux_f8  <= ramrdt_mux_f7;

				if (ramrdt_mux(2 downto 0) = "000" and ramrdt_mux_f1(2 downto 0) = "000") then
					col_b_f2 <= '1';
				else
					col_b_f2 <= '0';
				end if;

				if (ramrdt_mux = "1111" and ramrdt_mux_f1 = "1111") then
					col_w_f2 <= '1';
				else
					col_w_f2 <= '0';
				end if;

				col_b_f3       <= col_b_f2;
				col_b_f4       <= col_b_f3;
				col_b_f5       <= col_b_f4;
				col_b_f6       <= col_b_f5;
				col_b_f7       <= col_b_f6;
				col_b_f8       <= col_b_f7;
				col_b_f9       <= col_b_f8;
				col_b_f10      <= col_b_f9;
				col_w_f3       <= col_w_f2;
				col_w_f4       <= col_w_f3;
				col_w_f5       <= col_w_f4;
				col_w_f6       <= col_w_f5;
				col_w_f7       <= col_w_f6;
				col_w_f8       <= col_w_f7;
				col_w_f9       <= col_w_f8;
				col_w_f10      <= col_w_f9;

				r_hvld2_f1     <= r_hvld_f2;
				r_hvld2_f2     <= r_hvld2_f1;
				r_hvld2_f3     <= r_hvld2_f2;
				r_hvld2_f4     <= r_hvld2_f3;
				r_hvld2_f5     <= r_hvld2_f4;
				r_hvld2_f6     <= r_hvld2_f5;
				r_hvld2_f7     <= r_hvld2_f6;
				r_hvld2_f8     <= r_hvld2_f7;

			end if;
		end if;
	end process;


-- SC4COLORON    = 0/1 SCREEN4 color off/on
-- SC4COLORMD(0) = 0/1 color normal/reverse
-- SC4COLORMD(1) = 0/1 GREEN-PINK / RED-BLUE


--   Even						/		/
--   v
-- x000x 						X"000"		X"000"
-- x001x ɂ݈×				X"000""196"	X"000""159"
-- x010x ɂݓ				X"F0E"		X"F30"
-- x011x 						X"EFD"		X"EFD"
-- x100x ɂ݈×				X"196""000"	X"159""000"
-- 01010 ɂݗ				X"2F5"		X"2CF"
-- 01011 ɂ݈×				X"196""000"	X"159""000"
-- 11010 ɂ݈×				X"000""196"	X"000""159"
-- 11011 ɂ݈×				X""196"		X"159"
-- x110x 						X"EFD"		X"EFD"
-- x111x 						X"EFD"		X"EFD"
-- 43210

--   Odd						/		/
--   v
-- x000x 						X"000"		X"000"
-- x001x ɂ݈Ó				X"000""906"	X"000""920"
-- x010x ɂݗ				X"2F5"		X"2CF"
-- x011x 						X"EFD"		X"EFD"
-- x100x ɂ݈Ó				X"906""000"	X"920""000"
-- 01010 ɂݓ				X"F0E"		X"F30"
-- 01011 ɂ݈Ó				X"906""000"	X"920""000"
-- 11010 ɂ݈Ó				X"000""906"	X"000""920"
-- 11011 ɂ݈Ó				X"906"		X"920"
-- x110x 						X"EFD"		X"EFD"
-- x111x 						X"EFD"		X"EFD"
-- 43210


	bitpat <=
		col_w_f10 & col_w_f8 & col_w_f6 & col_w_f4 & col_w_f2;

	allsc4 <=
			'0' when (col_w_f10 = '0' and col_b_f10 = '0') else
			'0' when (col_w_f8  = '0' and col_b_f8  = '0') else
			'0' when (col_w_f6  = '0' and col_b_f6  = '0') else
			'0' when (col_w_f4  = '0' and col_b_f4  = '0') else
			'0' when (col_w_f2  = '0' and col_b_f2  = '0') else
			'1';


	process (CLK108M,RSTN)
	begin
		if (RSTN = '0') then
			bitpat_f1 <= (others => '0');
			allsc4_f1 <= '0';
		elsif (CLK108M'event and CLK108M = '1') then
			if (hcntrdtmp(0 downto 0) = "1") then
				bitpat_f1 <= bitpat;
				allsc4_f1 <= allsc4;
			end if;
		end if;
	end process;

	bitpatsel <= bitpat when (hcntrdtmp(1) = '0') else bitpat_f1;
	allsc4sel <= allsc4 when (hcntrdtmp(1) = '0') else allsc4_f1;


	c_dg <= X"1E9764";	-- ɂ݈×
	c_g  <= X"26FF5C";	-- ɂݗ
	c_dp <= X"970069";	-- ɂ݈Ó
	c_p  <= X"FF00EC";	-- ɂݓ
	c_db <= X"1E5096";	-- ɂ݈Ð
	c_b  <= X"26C9FF";	-- ɂݐ
	c_dr <= X"962D00";	-- ɂ݈Ð
	c_r  <= X"FF3000";	-- ɂݐ
	c_w  <= X"E0FFD0";	-- ɂݔ
	c_00 <= X"000000";	-- 


	set2p <=
		c_00 when (allsc4sel = '0') else
		c_00 when (SC4COLORON = '0') else
		c_00 when (SC4COLORMD(1) = '0') else
		c_00 when (bitpatsel(3 downto 1) = "000") else

		c_00 when (bitpatsel(3 downto 1) = "001"  and hcntrdtmp(2) = '0') else
		c_00 when (bitpatsel(3 downto 1) = "001"  and hcntrdtmp(2 downto 1) = "10") else
		c_dg when (bitpatsel(3 downto 1) = "001"  and hcntrdtmp(2 downto 1) = "11") else

		c_p  when (bitpatsel(3 downto 1) = "010") else
		c_w  when (bitpatsel(3 downto 1) = "011") else

		c_dg when (bitpatsel(3 downto 1) = "100"  and hcntrdtmp(2 downto 1) = "00") else
		c_00 when (bitpatsel(3 downto 1) = "100"  and hcntrdtmp(2 downto 1) = "01") else
		c_00 when (bitpatsel(3 downto 1) = "100"  and hcntrdtmp(2) = '1') else

		c_g  when (bitpatsel            = "01010") else

		c_00 when (bitpatsel            = "01011" and hcntrdtmp(2 downto 1) = "00") else
		c_dg when (bitpatsel            = "01011" and hcntrdtmp(2 downto 1) = "01") else
		c_00 when (bitpatsel            = "01011" and hcntrdtmp(2) = '1') else

		c_00 when (bitpatsel            = "11010" and hcntrdtmp(2) = '0') else
		c_dg when (bitpatsel            = "11010" and hcntrdtmp(2 downto 1) = "10") else
		c_00 when (bitpatsel            = "11010" and hcntrdtmp(2 downto 1) = "11") else

		c_dg when (bitpatsel            = "11011") else
		c_w  when (bitpatsel(3 downto 1) = "110") else
		c_w  when (bitpatsel(3 downto 1) = "111") else
		c_00;

	set2g <=
		c_00 when (allsc4sel = '0') else
		c_00 when (SC4COLORON = '0') else
		c_00 when (SC4COLORMD(1) = '0') else
		c_00 when (bitpatsel(3 downto 1) = "000") else

		c_00 when (bitpatsel(3 downto 1) = "001"  and hcntrdtmp(2) = '0') else
		c_00 when (bitpatsel(3 downto 1) = "001"  and hcntrdtmp(2 downto 1) = "10") else
		c_dp when (bitpatsel(3 downto 1) = "001"  and hcntrdtmp(2 downto 1) = "11") else

		c_g  when (bitpatsel(3 downto 1) = "010") else
		c_w  when (bitpatsel(3 downto 1) = "011") else

		c_dp when (bitpatsel(3 downto 1) = "100"  and hcntrdtmp(2 downto 1) = "00") else
		c_00 when (bitpatsel(3 downto 1) = "100"  and hcntrdtmp(2 downto 1) = "01") else
		c_00 when (bitpatsel(3 downto 1) = "100"  and hcntrdtmp(2) = '1') else

		c_p  when (bitpatsel            = "01010") else

		c_00 when (bitpatsel            = "01011" and hcntrdtmp(2 downto 1) = "00") else
		c_dp when (bitpatsel            = "01011" and hcntrdtmp(2 downto 1) = "01") else
		c_00 when (bitpatsel            = "01011" and hcntrdtmp(2) = '1') else

		c_00 when (bitpatsel            = "11010" and hcntrdtmp(2) = '0') else
		c_dp when (bitpatsel            = "11010" and hcntrdtmp(2 downto 1) = "10") else
		c_00 when (bitpatsel            = "11010" and hcntrdtmp(2 downto 1) = "11") else

		c_dp when (bitpatsel            = "11011") else
		c_w  when (bitpatsel(3 downto 1) = "110") else
		c_w  when (bitpatsel(3 downto 1) = "111") else
		c_00;



	set2r <=
		c_00 when (allsc4sel = '0') else
		c_00 when (SC4COLORON = '0') else
		c_00 when (SC4COLORMD(1) = '1') else
		c_00 when (bitpatsel(3 downto 1) = "000") else

		c_00 when (bitpatsel(3 downto 1) = "001"  and hcntrdtmp(2) = '0') else
		c_00 when (bitpatsel(3 downto 1) = "001"  and hcntrdtmp(2 downto 1) = "10") else
		c_db when (bitpatsel(3 downto 1) = "001"  and hcntrdtmp(2 downto 1) = "11") else

		c_r  when (bitpatsel(3 downto 1) = "010") else
		c_w  when (bitpatsel(3 downto 1) = "011") else

		c_db when (bitpatsel(3 downto 1) = "100"  and hcntrdtmp(2 downto 1) = "00") else
		c_00 when (bitpatsel(3 downto 1) = "100"  and hcntrdtmp(2 downto 1) = "01") else
		c_00 when (bitpatsel(3 downto 1) = "100"  and hcntrdtmp(2) = '1') else

		c_b  when (bitpatsel            = "01010") else

		c_00 when (bitpatsel            = "01011" and hcntrdtmp(2 downto 1) = "00") else
		c_db when (bitpatsel            = "01011" and hcntrdtmp(2 downto 1) = "01") else
		c_00 when (bitpatsel            = "01011" and hcntrdtmp(2) = '1') else

		c_00 when (bitpatsel            = "11010" and hcntrdtmp(2) = '0') else
		c_db when (bitpatsel            = "11010" and hcntrdtmp(2 downto 1) = "10") else
		c_00 when (bitpatsel            = "11010" and hcntrdtmp(2 downto 1) = "11") else

		c_db when (bitpatsel            = "11011") else
		c_w  when (bitpatsel(3 downto 1) = "110") else
		c_w  when (bitpatsel(3 downto 1) = "111") else
		c_00;

	set2b <=
		c_00 when (allsc4sel = '0') else
		c_00 when (SC4COLORON = '0') else
		c_00 when (SC4COLORMD(1) = '1') else
		c_00 when (bitpatsel(3 downto 1) = "000") else

		c_00 when (bitpatsel(3 downto 1) = "001"  and hcntrdtmp(2) = '0') else
		c_00 when (bitpatsel(3 downto 1) = "001"  and hcntrdtmp(2 downto 1) = "10") else
		c_dr when (bitpatsel(3 downto 1) = "001"  and hcntrdtmp(2 downto 1) = "11") else

		c_b  when (bitpatsel(3 downto 1) = "010") else
		c_w  when (bitpatsel(3 downto 1) = "011") else

		c_dr when (bitpatsel(3 downto 1) = "100"  and hcntrdtmp(2 downto 1) = "00") else
		c_00 when (bitpatsel(3 downto 1) = "100"  and hcntrdtmp(2 downto 1) = "01") else
		c_00 when (bitpatsel(3 downto 1) = "100"  and hcntrdtmp(2) = '1') else

		c_r  when (bitpatsel            = "01010") else

		c_00 when (bitpatsel            = "01011" and hcntrdtmp(2 downto 1) = "00") else
		c_dr when (bitpatsel            = "01011" and hcntrdtmp(2 downto 1) = "01") else
		c_00 when (bitpatsel            = "01011" and hcntrdtmp(2) = '1') else

		c_00 when (bitpatsel            = "11010" and hcntrdtmp(2) = '0') else
		c_dr when (bitpatsel            = "11010" and hcntrdtmp(2 downto 1) = "10") else
		c_00 when (bitpatsel            = "11010" and hcntrdtmp(2 downto 1) = "11") else

		c_dr when (bitpatsel            = "11011") else
		c_w  when (bitpatsel(3 downto 1) = "110") else
		c_w  when (bitpatsel(3 downto 1) = "111") else
		c_00;


	process (CLK108M,RSTN)
	begin
		if (RSTN = '0') then
			set2p_f1 <= (others => '0');
			set2p_f2 <= (others => '0');
			set2r_f1 <= (others => '0');
			set2r_f2 <= (others => '0');
			set2g_f1 <= (others => '0');
			set2g_f2 <= (others => '0');
			set2b_f1 <= (others => '0');
			set2b_f2 <= (others => '0');
		elsif (CLK108M'event and CLK108M = '1') then
			if (hcntrdtmp(0 downto 0) = "1") then
				set2p_f1 <= set2p;
				set2p_f2 <= set2p_f1;
				set2r_f1 <= set2r;
				set2r_f2 <= set2r_f1;
				set2g_f1 <= set2g;
				set2g_f2 <= set2g_f1;
				set2b_f1 <= set2b;
				set2b_f2 <= set2b_f1;
			end if;
		end if;
	end process;

	process (CLK108M,RSTN)
	begin
		if (RSTN = '0') then
			r_hvld3_f1 <= '0';
			r_hvld3_f2 <= '0';
		elsif (CLK108M'event and CLK108M = '1') then
			r_hvld3_f1 <= r_hvld2_f8;
			r_hvld3_f2 <= r_hvld3_f1;
		end if;
	end process;

	r_hvld2_sel    <= r_hvld3_f2;
	ramrdt_mux_sel <= ramrdt_mux_f8;

	vgadat <=
		X"000000" when (r_hvld2_sel = '0') else

		X"1E5096" when (DT_LOCK3 = '0' and DT_LOCK4 = '0') else		-- no sync

		set2p_f2  when (set2p_f2 /= X"000000" and hcntrdtmp(2) =     SC4COLORMD(0) ) else
		set2r_f2  when (set2r_f2 /= X"000000" and hcntrdtmp(2) =     SC4COLORMD(0) ) else
		set2g_f2  when (set2g_f2 /= X"000000" and hcntrdtmp(2) = not SC4COLORMD(0) ) else
		set2b_f2  when (set2b_f2 /= X"000000" and hcntrdtmp(2) = not SC4COLORMD(0) ) else

		X"000000" when (ramrdt_mux_sel = "0000") else	-- mk2 ()
		X"B000FF" when (ramrdt_mux_sel = "0001") else	-- mk2 
		X"FFB000" when (ramrdt_mux_sel = "0010") else	-- mk2 
		X"FF00B0" when (ramrdt_mux_sel = "0011") else	-- mk2 Ԏ
		X"00FFB0" when (ramrdt_mux_sel = "0100") else	-- mk2 
		X"00B0FF" when (ramrdt_mux_sel = "0101") else	-- mk2 F
		X"B0FF00" when (ramrdt_mux_sel = "0110") else	-- mk2 
		X"B0B0B0" when (ramrdt_mux_sel = "0111") else	-- mk2 DF
		X"000000" when (ramrdt_mux_sel = "1000") else	-- mk2 
		X"0000FF" when (ramrdt_mux_sel = "1001") else	-- mk2 
		X"FF0000" when (ramrdt_mux_sel = "1010") else	-- mk2 
		X"FF00FF" when (ramrdt_mux_sel = "1011") else	-- mk2 }[^
		X"00FF00" when (ramrdt_mux_sel = "1100") else	-- mk2 
		X"00FFFF" when (ramrdt_mux_sel = "1101") else	-- mk2 VA
		X"FFFF00" when (ramrdt_mux_sel = "1110") else	-- mk2 
		X"FFFFFF" when (ramrdt_mux_sel = "1111") else	-- mk2 

		X"000000";

	vga_hs_sel <= vga_hs_f10;
	vga_vs_sel <= vga_vs_f10;


	process (CLK108M,RSTN)
	begin
		if (RSTN = '0') then
			vga_r_o  <= (others => '0');
			vga_g_o  <= (others => '0');
			vga_b_o  <= (others => '0');
			vga_hs_o <= '1';
			vga_vs_o <= '1';
		elsif (CLK108M'event and CLK108M = '1') then
			vga_r_o  <= vgadat(23 downto 16);
			vga_g_o  <= vgadat(15 downto  8);
			vga_b_o  <= vgadat( 7 downto  0);
			vga_hs_o <= vga_hs_sel;
			vga_vs_o <= vga_vs_sel;
		end if;
	end process;

	VGA_R	<= vga_r_o(7 downto 4);
	VGA_G	<= vga_g_o(7 downto 4);
	VGA_B	<= vga_b_o(7 downto 4);
	VGA_HS	<= vga_hs_o;
	VGA_VS	<= vga_vs_o;

end RTL;
